"""Python library to interact with UniFi controller."""

from http import HTTPStatus
import logging
from pprint import pformat
from ssl import SSLContext
from typing import Any, Callable, Dict, List, Optional, Union

import aiohttp
from aiohttp import client_exceptions

from .clients import Clients, ClientsAll
from .devices import Devices
from .dpi import DPIRestrictionApps, DPIRestrictionGroups
from .errors import (
    BadGateway,
    LoginRequired,
    RequestError,
    ResponseError,
    ServiceUnavailable,
    raise_error,
)
from .events import CLIENT_EVENTS, DEVICE_EVENTS, Event
from .websocket import SIGNAL_CONNECTION_STATE, SIGNAL_DATA, WSClient
from .wlan import Wlans

LOGGER = logging.getLogger(__name__)

MESSAGE_CLIENT = "sta:sync"
MESSAGE_CLIENT_REMOVED = "user:delete"
MESSAGE_DEVICE = "device:sync"
MESSAGE_EVENT = "events"
MESSAGE_DPI_GROUP = "dpigroup"
MESSAGE_DPI_APP = "dpiapp"

ATTR_MESSAGE = "message"
ATTR_META = "meta"
ATTR_DATA = "data"

DATA_CLIENT = "client"
DATA_CLIENT_REMOVED = "client_removed"
DATA_DEVICE = "device"
DATA_EVENT = "event"
DATA_DPI_APP = "dpi_app"
DATA_DPI_APP_REMOVED = "dpi_app_removed"
DATA_DPI_GROUP = "dpi_group"
DATA_DPI_GROUP_REMOVED = "dpi_group_removed"


IGNORE_MESSAGES = ("device:update",)


class Controller:
    """Control a UniFi controller."""

    def __init__(
        self,
        host: str,
        websession: aiohttp.ClientSession,
        *,
        username: str,
        password: str,
        port=8443,
        site="default",
        sslcontext: Optional[SSLContext] = None,
        callback: Optional[Callable[[str, Union[dict, str]], None]] = None,
    ):
        """Session setup."""
        self.host = host
        self.session = websession
        self.port = port
        self.username = username
        self.password = password
        self.site = site
        self.sslcontext = sslcontext
        self.callback = callback
        self.can_retry_login = False

        self.url = f"https://{self.host}:{self.port}"
        self.is_unifi_os = False
        self.headers: Dict[str, Any] = {}
        self.last_response: Optional[aiohttp.ClientResponse] = None

        self.websocket: Optional[WSClient] = None

        self.clients = Clients([], self.request)
        self.devices = Devices([], self.request)
        self.dpi_apps = DPIRestrictionApps([], self.request)
        self.dpi_groups = DPIRestrictionGroups([], self.request, self.dpi_apps)
        self.clients_all = ClientsAll([], self.request)
        self.wlans = Wlans([], self.request)

    async def check_unifi_os(self) -> None:
        """Check if controller is running UniFi OS."""
        await self._request("get", url=self.url, allow_redirects=False)
        if (
            response := self.last_response
        ) is not None and response.status == HTTPStatus.OK:
            self.is_unifi_os = True
            self.headers = {"x-csrf-token": response.headers.get("x-csrf-token")}

    async def login(self) -> None:
        """Log in to controller."""
        if self.is_unifi_os:
            url = f"{self.url}/api/auth/login"
        else:
            url = f"{self.url}/api/login"

        auth = {
            "username": self.username,
            "password": self.password,
            "remember": True,
        }

        await self._request("post", url=url, json=auth)

        self.can_retry_login = True

    async def sites(self) -> dict:
        """Retrieve what sites are provided by controller."""
        if self.is_unifi_os:
            url = f"{self.url}/proxy/network/api/self/sites"
        else:
            url = f"{self.url}/api/self/sites"

        sites = await self.request("get", url=url)
        LOGGER.debug(pformat(sites))
        return {site["desc"]: site for site in sites}

    async def site_description(self) -> List[dict]:
        """User description of current site."""
        description = await self.request("get", "/self")
        LOGGER.debug(description)
        return description

    async def initialize(self) -> None:
        """Load UniFi parameters."""
        await self.clients.update()
        await self.clients_all.update()
        await self.devices.update()
        await self.dpi_apps.update()
        await self.dpi_groups.update()
        await self.wlans.update()

    def start_websocket(self) -> None:
        """Start websession and websocket to UniFi."""
        self.websocket = WSClient(
            self.session,
            self.host,
            self.port,
            self.sslcontext,
            self.site,
            callback=self.session_handler,
            is_unifi_os=self.is_unifi_os,
        )
        self.websocket.start()

    def stop_websocket(self) -> None:
        """Close websession and websocket to UniFi."""
        LOGGER.info("Shutting down connections to UniFi.")
        if self.websocket:
            self.websocket.stop()

    def session_handler(self, signal: str) -> None:
        """Signalling from websocket.

        data - new data available for processing.
        state - network state has changed.
        """
        if not self.websocket:
            return

        if signal == SIGNAL_DATA:
            new_items = self.message_handler(self.websocket.data)
            if new_items and self.callback:
                self.callback(SIGNAL_DATA, new_items)

        elif signal == SIGNAL_CONNECTION_STATE and self.callback:
            self.callback(SIGNAL_CONNECTION_STATE, self.websocket.state)

    def message_handler(self, message: dict) -> dict:
        """Receive event from websocket and identifies where the event belong."""
        changes = {}

        if message[ATTR_META][ATTR_MESSAGE] == MESSAGE_CLIENT:
            changes[DATA_CLIENT] = self.clients.process_raw(message[ATTR_DATA])

        elif message[ATTR_META][ATTR_MESSAGE] == MESSAGE_DEVICE:
            changes[DATA_DEVICE] = self.devices.process_raw(message[ATTR_DATA])

        elif message[ATTR_META][ATTR_MESSAGE] == MESSAGE_EVENT:
            events = [Event(raw_event) for raw_event in message[ATTR_DATA]]
            self.clients.process_event(
                [
                    client_event
                    for client_event in events
                    if client_event.event in CLIENT_EVENTS
                ]
            )
            self.devices.process_event(
                [
                    device_event
                    for device_event in events
                    if device_event.event in DEVICE_EVENTS
                ]
            )
            changes[DATA_EVENT] = set(events)

        elif message[ATTR_META][ATTR_MESSAGE] == MESSAGE_CLIENT_REMOVED:
            changes[DATA_CLIENT_REMOVED] = self.clients.remove(message[ATTR_DATA])

        elif MESSAGE_DPI_GROUP in message[ATTR_META][ATTR_MESSAGE]:
            action = message[ATTR_META][ATTR_MESSAGE].split(":")[1]
            if action == "delete":
                changes[DATA_DPI_GROUP_REMOVED] = self.dpi_groups.remove(
                    message[ATTR_DATA]
                )
            else:
                changes[DATA_DPI_GROUP] = self.dpi_groups.process_raw(
                    message[ATTR_DATA]
                )

        elif MESSAGE_DPI_APP in message[ATTR_META][ATTR_MESSAGE]:
            action = message[ATTR_META][ATTR_MESSAGE].split(":")[1]
            if action == "delete":
                changes[DATA_DPI_APP_REMOVED] = self.dpi_apps.remove(message[ATTR_DATA])
            else:
                changes[DATA_DPI_APP] = self.dpi_apps.process_raw(message[ATTR_DATA])
                if action == "sync":
                    # Manually trigger update for related groups
                    changes[DATA_DPI_GROUP] = {
                        key
                        for key in self.dpi_groups
                        if set(self.dpi_groups[key].dpiapp_ids).intersection(
                            changes[DATA_DPI_APP]
                        )
                    }
                    for key in changes[DATA_DPI_GROUP]:
                        group = self.dpi_groups[key]
                        group.update(group.raw)

        elif message[ATTR_META][ATTR_MESSAGE] in IGNORE_MESSAGES:
            pass

        else:
            LOGGER.debug("Unsupported message type %s", message)

        return changes

    async def request(
        self,
        method: str,
        path: str = "",
        json: Optional[Dict[str, Any]] = None,
        url: str = "",
    ) -> List[dict]:
        """Make a request to the API, retry login on failure."""
        try:
            return await self._request(method, path, json, url)
        except LoginRequired:
            if not self.can_retry_login:
                raise
            # Session likely expired, try again
            self.can_retry_login = False
            # Make sure we get a new csrf token
            await self.check_unifi_os()
            await self.login()
            return await self._request(method, path, json, url)

    async def _request(
        self,
        method: str,
        path: str = "",
        json: Optional[Dict[str, Any]] = None,
        url: str = "",
        **kwargs: bool,
    ) -> List[dict]:
        """Make a request to the API."""
        self.last_response = None

        if not url:
            if self.is_unifi_os:
                url = f"{self.url}/proxy/network/api/s/{self.site}"
            else:
                url = f"{self.url}/api/s/{self.site}"

            url += f"{path}"

        LOGGER.debug("%s", url)

        try:
            async with self.session.request(
                method,
                url,
                json=json,
                ssl=self.sslcontext,
                headers=self.headers,
                **kwargs,
            ) as res:
                LOGGER.debug("%s %s %s", res.status, res.content_type, res)

                self.last_response = res

                if res.status == HTTPStatus.UNAUTHORIZED:
                    raise LoginRequired(f"Call {url} received 401 Unauthorized")

                if res.status == HTTPStatus.NOT_FOUND:
                    raise ResponseError(f"Call {url} received 404 Not Found")

                if res.status == HTTPStatus.BAD_GATEWAY:
                    raise BadGateway(f"Call {url} received 502 bad gateway")

                if res.status == HTTPStatus.SERVICE_UNAVAILABLE:
                    raise ServiceUnavailable(
                        f"Call {url} received 503 service unavailable"
                    )

                if res.content_type == "application/json":
                    response = await res.json()
                    _raise_on_error(response)
                    if "data" in response:
                        return response["data"]
                    return response
                return []

        except client_exceptions.ClientError as err:
            raise RequestError(
                f"Error requesting data from {self.host}: {err}"
            ) from None


def _raise_on_error(data) -> None:
    """Check response for error message."""
    if not isinstance(data, dict):
        return None

    if "meta" in data and data["meta"]["rc"] == "error":
        raise_error(data["meta"]["msg"])

    if "errors" in data:
        raise_error(data["errors"][0])



"""Clients are devices on a UniFi network."""

from typing import Awaitable, Callable, List, Optional

from .api import APIItem, APIItems

URL = "/stat/sta"  # Active clients
URL_ALL = "/rest/user"  # All known and configured clients

URL_CLIENT_STATE_MANAGER = "/cmd/stamgr"


class Clients(APIItems):
    """Represents client network devices."""

    KEY = "mac"

    def __init__(
        self,
        raw: List[dict],
        request: Callable[..., Awaitable[List[dict]]],
    ) -> None:
        """Initialize active clients manager."""
        super().__init__(raw, request, URL, Client)

    async def async_block(self, mac: str) -> List[dict]:
        """Block client from controller."""
        data = {"mac": mac, "cmd": "block-sta"}
        return await self._request("post", URL_CLIENT_STATE_MANAGER, json=data)

    async def async_unblock(self, mac: str) -> List[dict]:
        """Unblock client from controller."""
        data = {"mac": mac, "cmd": "unblock-sta"}
        return await self._request("post", URL_CLIENT_STATE_MANAGER, json=data)

    async def async_reconnect(self, mac: str) -> List[dict]:
        """Force a wireless client to reconnect to the network."""
        data = {"mac": mac, "cmd": "kick-sta"}
        return await self._request("post", URL_CLIENT_STATE_MANAGER, json=data)

    async def remove_clients(self, macs: List[str]) -> List[dict]:
        """Make controller forget provided clients."""
        data = {"macs": macs, "cmd": "forget-sta"}
        return await self._request("post", URL_CLIENT_STATE_MANAGER, json=data)


class ClientsAll(APIItems):
    """Represents all client network devices."""

    KEY = "mac"

    def __init__(
        self,
        raw: List[dict],
        request: Callable[..., Awaitable[List[dict]]],
    ) -> None:
        """Initialize all clients manager."""
        super().__init__(raw, request, URL_ALL, Client)


class Client(APIItem):
    """Represents a client network device."""

    @property
    def access_point_mac(self) -> str:
        """MAC address of access point."""
        return self.raw.get("ap_mac", "")

    @property
    def association_time(self) -> Optional[int]:
        """When was client associated with controller."""
        return self.raw.get("assoc_time")

    @property
    def blocked(self) -> bool:
        """Is client blocked."""
        return self.raw.get("blocked", False)

    @property
    def device_name(self) -> str:
        """Device name of client."""
        return self.raw.get("device_name", "")

    @property
    def essid(self) -> str:
        """ESSID client is connected to."""
        return self.raw.get("essid", "")

    @property
    def first_seen(self) -> Optional[int]:
        """When was client first seen."""
        return self.raw.get("first_seen")

    @property
    def fixed_ip(self) -> str:
        """List IP if fixed IP is configured."""
        return self.raw.get("fixed_ip", "")

    @property
    def hostname(self) -> str:
        """Hostname of client."""
        return self.raw.get("hostname", "")

    @property
    def ip(self) -> str:
        """IP of client."""
        return self.raw.get("ip", "")

    @property
    def is_guest(self) -> bool:
        """Is client guest."""
        return self.raw.get("is_guest", False)

    @property
    def is_wired(self) -> bool:
        """Is client wired."""
        return self.raw.get("is_wired", False)

    @property
    def last_seen(self) -> Optional[int]:
        """When was client last seen."""
        return self.raw.get("last_seen")

    @property
    def latest_association_time(self) -> Optional[int]:
        """When was client last associated with controller."""
        return self.raw.get("latest_assoc_time")

    @property
    def mac(self) -> str:
        """MAC address of client."""
        return self.raw.get("mac", "")

    @property
    def name(self) -> str:
        """Name of client."""
        return self.raw.get("name", "")

    @property
    def oui(self) -> str:
        """Vendor string for client MAC."""
        return self.raw.get("oui", "")

    @property
    def site_id(self) -> str:
        """Site ID client belongs to."""
        return self.raw.get("site_id", "")

    @property
    def sw_depth(self) -> Optional[int]:
        """How many layers of switches client is in."""
        return self.raw.get("sw_depth")

    @property
    def sw_mac(self) -> str:
        """MAC for switch client is connected to."""
        return self.raw.get("sw_mac", "")

    @property
    def sw_port(self) -> Optional[int]:
        """Switch port client is connected to."""
        return self.raw.get("sw_port")

    @property
    def rx_bytes(self) -> int:
        """Bytes received over wireless connection."""
        return self.raw.get("rx_bytes", 0)

    @property
    def rx_bytes_r(self) -> int:
        """Bytes recently received over wireless connection."""
        return self.raw.get("rx_bytes-r", 0)

    @property
    def tx_bytes(self) -> int:
        """Bytes transferred over wireless connection."""
        return self.raw.get("tx_bytes", 0)

    @property
    def tx_bytes_r(self) -> int:
        """Bytes recently transferred over wireless connection."""
        return self.raw.get("tx_bytes-r", 0)

    @property
    def uptime(self) -> int:
        """Uptime of client."""
        return self.raw.get("uptime", 0)

    @property
    def wired_rx_bytes(self) -> int:
        """Bytes received over wired connection."""
        return self.raw.get("wired-rx_bytes", 0)

    @property
    def wired_rx_bytes_r(self) -> int:
        """Bytes recently received over wired connection."""
        return self.raw.get("wired-rx_bytes-r", 0)

    @property
    def wired_tx_bytes(self) -> int:
        """Bytes transferred over wired connection."""
        return self.raw.get("wired-tx_bytes", 0)

    @property
    def wired_tx_bytes_r(self) -> int:
        """Bytes recently transferred over wired connection."""
        return self.raw.get("wired-tx_bytes-r", 0)

    def __repr__(self) -> str:
        """Return the representation."""
        name = self.name or self.hostname
        return f"<Client {name}: {self.mac} {self.raw}>"



"""Python library to connect UniFi and Home Assistant to work together."""

import asyncio
import json
import logging
from ssl import SSLContext
from typing import Callable, Optional

import aiohttp

LOGGER = logging.getLogger(__name__)

SIGNAL_DATA = "data"
SIGNAL_CONNECTION_STATE = "state"

STATE_DISCONNECTED = "disconnected"
STATE_RUNNING = "running"
STATE_STARTING = "starting"
STATE_STOPPED = "stopped"


class WSClient:
    """Websocket transport, session handling, message generation."""

    def __init__(
        self,
        session: aiohttp.ClientSession,
        host: str,
        port: int,
        ssl_context: Optional[SSLContext],
        site: str,
        callback: Callable[[str], None],
        is_unifi_os: bool = False,
    ):
        """Create resources for websocket communication."""
        self.session = session
        self.ssl_context = ssl_context
        self.session_handler_callback = callback

        if is_unifi_os:
            self.url = f"wss://{host}:{port}/proxy/network/wss/s/{site}/events"
        else:
            self.url = f"wss://{host}:{port}/wss/s/{site}/events"

        self._loop = asyncio.get_running_loop()

        self._data: dict = {}
        self._state = ""

    @property
    def data(self) -> dict:
        """Return data."""
        return self._data

    @property
    def state(self) -> str:
        """State of websocket."""
        return self._state

    @state.setter
    def state(self, value: str) -> None:
        """Set state of websocket."""
        self._state = value
        LOGGER.debug("Websocket %s", value)
        self.session_handler_callback(SIGNAL_CONNECTION_STATE)

    def start(self) -> None:
        """Start websocket and update its state."""
        if self.state != STATE_RUNNING:
            self.state = STATE_STARTING
            self._loop.create_task(self.running())

    def stop(self) -> None:
        """Close websocket connection."""
        self.state = STATE_STOPPED

    async def running(self) -> None:
        """Start websocket connection."""
        try:
            async with self.session.ws_connect(
                self.url, ssl=self.ssl_context, heartbeat=15
            ) as ws:
                self.state = STATE_RUNNING

                async for msg in ws:

                    if self.state == STATE_STOPPED:
                        break

                    if msg.type == aiohttp.WSMsgType.TEXT:
                        self._data = json.loads(msg.data)
                        self.session_handler_callback(SIGNAL_DATA)
                        LOGGER.debug(msg.data)

                    elif msg.type == aiohttp.WSMsgType.CLOSED:
                        LOGGER.warning("AIOHTTP websocket connection closed")
                        break

                    elif msg.type == aiohttp.WSMsgType.ERROR:
                        LOGGER.error("AIOHTTP websocket error")
                        break

        except aiohttp.ClientConnectorError:
            if self.state != STATE_STOPPED:
                LOGGER.error("Client connection error")
                self.state = STATE_DISCONNECTED

        except Exception as err:
            if self.state != STATE_STOPPED:
                LOGGER.error("Unexpected error %s", err)
                self.state = STATE_DISCONNECTED

        else:
            if self.state != STATE_STOPPED:
                self.state = STATE_DISCONNECTED



"""DPI Restrictions as part of a UniFi network."""
import asyncio
from typing import Awaitable, Callable, List, Optional

from .api import APIItem, APIItems

GROUP_URL = "/rest/dpigroup"  # List DPI Group configuration
APP_URL = "/rest/dpiapp"  # List DPI App configuration


class DPIRestrictionApp(APIItem):
    """Represents a DPI App configuration."""

    @property
    def id(self) -> str:
        """DPI app ID."""
        return self.raw["_id"]

    @property
    def apps(self) -> list:
        """List of apps."""
        return self.raw["apps"]

    @property
    def blocked(self) -> bool:
        """Is blocked."""
        return self.raw["blocked"]

    @property
    def cats(self) -> list:
        """Categories."""
        return self.raw["cats"]

    @property
    def enabled(self) -> bool:
        """Is enabled."""
        return self.raw["enabled"]

    @property
    def log(self) -> bool:
        """Is logging enabled."""
        return self.raw["log"]

    @property
    def site_id(self) -> str:
        """Site ID."""
        return self.raw["site_id"]


class DPIRestrictionApps(APIItems):
    """Represents DPI App configurations."""

    KEY = "_id"

    def __init__(
        self,
        raw: List[dict],
        request: Callable[..., Awaitable[List[dict]]],
    ) -> None:
        """Initialize DPI restriction apps manager."""
        super().__init__(raw, request, APP_URL, DPIRestrictionApp)

    async def async_enable(self, app_id: str) -> List[dict]:
        """Enable DPI Restriction Group Apps."""
        app_url = f"{APP_URL}/{app_id}"
        data = {"enabled": True}
        return await self._request("put", app_url, json=data)

    async def async_disable(self, app_id: str) -> List[dict]:
        """Disable DPI Restriction Group Apps."""
        app_url = f"{APP_URL}/{app_id}"
        data = {"enabled": False}
        return await self._request("put", app_url, json=data)

    def process_raw(self, raw: list) -> set:
        """Set enabled attribute based on app data."""
        super().process_raw(raw)

        return {raw_item[self.KEY] for raw_item in raw}


class DPIRestrictionGroup(APIItem):
    """Represents a DPI Group configuration."""

    def __init__(
        self,
        raw: dict,
        request: Callable[..., Awaitable[List[dict]]],
    ) -> None:
        """Initialize DPI Group."""
        super().__init__(raw, request)
        self.apps: Optional[DPIRestrictionApps] = None

    @property
    def id(self) -> str:
        """DPI group ID."""
        return self.raw["_id"]

    @property
    def attr_no_delete(self) -> bool:
        """Can be deleted."""
        return self.raw.get("attr_no_delete", False)

    @property
    def attr_hidden_id(self) -> str:
        """Attr hidden ID."""
        return self.raw.get("attr_hidden_id", "")

    @property
    def name(self) -> str:
        """DPI group name."""
        return self.raw["name"]

    @property
    def site_id(self) -> str:
        """Site ID."""
        return self.raw["site_id"]

    @property
    def dpiapp_ids(self) -> List[str]:
        """DPI app IDs belonging to group."""
        return self.raw.get("dpiapp_ids", [])

    @property
    def enabled(self) -> bool:
        """Are all apps in group enabled."""
        return isinstance(self.apps, DPIRestrictionApps) and all(
            [
                app.enabled
                for app_id, app in self.apps.items()
                if app_id in self.dpiapp_ids
            ]
        )


class DPIRestrictionGroups(APIItems):
    """Represents DPI Group configurations."""

    KEY = "_id"

    def __init__(
        self,
        raw: list,
        request: Callable[..., Awaitable[List[dict]]],
        apps: DPIRestrictionApps,
    ) -> None:
        """Initialize DPI restriction group manager."""
        self._apps = apps
        super().__init__(raw, request, GROUP_URL, DPIRestrictionGroup)

    async def async_enable(self, group: DPIRestrictionGroup) -> List[dict]:
        """Enable DPI Restriction Group Apps."""
        return await asyncio.gather(
            *[self._apps.async_enable(app_id) for app_id in group.dpiapp_ids]
        )

    async def async_disable(self, group: DPIRestrictionGroup) -> List[dict]:
        """Disable DPI Restriction Group Apps."""
        return await asyncio.gather(
            *[self._apps.async_disable(app_id) for app_id in group.dpiapp_ids]
        )

    def process_raw(self, raw: list) -> set:
        """Set enabled attribute based on app data."""
        new_items = super().process_raw(raw)

        for group_id in new_items:
            group = self._items[group_id]
            group.apps = self._apps

        return new_items



"""UniFi devices are network infrastructure.

Access points, Gateways, Switches.
"""

import logging
from typing import (
    Awaitable,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Union,
    ValuesView,
)

from .api import APIItem, APIItems
from .events import Event as UniFiEvent

LOGGER = logging.getLogger(__name__)

URL = "/stat/device"


class Devices(APIItems):
    """Represents network devices."""

    KEY = "mac"

    def __init__(
        self,
        raw: List[dict],
        request: Callable[..., Awaitable[List[dict]]],
    ) -> None:
        """Initialize device manager."""
        super().__init__(raw, request, URL, Device)


class Device(APIItem):
    """Represents a network device."""

    def __init__(
        self,
        raw: dict,
        request: Callable[..., Awaitable[List[dict]]],
    ) -> None:
        """Initialize device."""
        super().__init__(raw, request)
        self.ports = Ports(raw.get("port_table", []))

    def update(
        self,
        raw: Optional[dict] = None,
        event: Optional[UniFiEvent] = None,
    ) -> None:
        """Refresh data."""
        if raw:
            self.ports.update(raw.get("port_table", []))
        super().update(raw, event)

    @property
    def board_rev(self) -> int:
        """Board revision of device."""
        return self.raw["board_rev"]

    @property
    def considered_lost_at(self) -> int:
        """When device is considered lost."""
        return self.raw["considered_lost_at"]

    @property
    def disabled(self) -> bool:
        """Is device disabled."""
        return self.raw.get("disabled", False)

    @property
    def id(self) -> str:
        """ID of device."""
        return self.raw["device_id"]

    @property
    def ip(self) -> str:
        """IP of device."""
        return self.raw["ip"]

    @property
    def fan_level(self) -> Optional[int]:
        """Fan level of device."""
        return self.raw.get("fan_level")

    @property
    def has_fan(self) -> bool:
        """Do device have a fan."""
        return self.raw.get("has_fan", False)

    @property
    def last_seen(self) -> Optional[int]:
        """When was device last seen."""
        return self.raw.get("last_seen")

    @property
    def mac(self) -> str:
        """MAC address of device."""
        return self.raw["mac"]

    @property
    def model(self) -> str:
        """Model of device."""
        return self.raw["model"]

    @property
    def name(self) -> str:
        """Name of device."""
        return self.raw.get("name", "")

    @property
    def next_heartbeat_at(self) -> Optional[int]:
        """Next heart beat full UNIX time."""
        return self.raw.get("next_heartbeat_at")

    @property
    def next_interval(self) -> int:
        """Next heart beat in seconds."""
        return self.raw.get("next_interval", 30)

    @property
    def overheating(self) -> bool:
        """Is device overheating."""
        return self.raw.get("overheating", False)

    @property
    def port_overrides(self) -> list:
        """Overridden port configuration."""
        return self.raw.get("port_overrides", [])

    @property
    def port_table(self) -> list:
        """List of ports and data."""
        return self.raw.get("port_table", [])

    @property
    def state(self) -> int:
        """State of device."""
        return self.raw["state"]

    @property
    def sys_stats(self) -> dict:
        """Output from top."""
        return self.raw["sys_stats"]

    @property
    def type(self) -> str:
        """Type of device."""
        return self.raw["type"]

    @property
    def version(self) -> str:
        """Firmware version."""
        return self.raw["version"]

    @property
    def upgradable(self) -> bool:
        """Is a new firmware available."""
        return self.raw.get("upgradable", False)

    @property
    def upgrade_to_firmware(self) -> str:
        """Firmware version to update to."""
        return self.raw.get("upgrade_to_firmware", "")

    @property
    def uplink_depth(self) -> Optional[int]:
        """Hops to gateway."""
        return self.raw.get("uplink_depth")

    @property
    def user_num_sta(self) -> int:
        """Amount of connected clients."""
        return self.raw["user-num_sta"]

    @property
    def wlan_overrides(self) -> list:
        """Wlan configuration override."""
        return self.raw.get("wlan_overrides", [])

    async def async_set_port_poe_mode(self, port_idx: int, mode: str) -> List[dict]:
        """Set port poe mode.

        Auto, 24v, passthrough, off.
        Make sure to not overwrite any existing configs.
        """
        LOGGER.debug("Override port %d with mode %s", port_idx, mode)

        existing_override = False
        for port_override in self.port_overrides:
            if port_idx == port_override["port_idx"]:
                port_override["poe_mode"] = mode
                existing_override = True
                break

        if not existing_override:
            self.port_overrides.append(
                {
                    "port_idx": port_idx,
                    "portconf_id": self.ports[port_idx].portconf_id,
                    "poe_mode": mode,
                }
            )

        url = f"/rest/device/{self.id}"
        data = {"port_overrides": self.port_overrides}

        return await self._request("put", url, json=data)

    def __repr__(self) -> str:
        """Return the representation."""
        return f"<Device {self.name}: {self.mac}>"


class Port:
    """Represents a network port."""

    def __init__(self, raw: dict) -> None:
        """Initialize port."""
        self.raw = raw

    @property
    def ifname(self) -> str:
        """Port name used by USG."""
        return self.raw.get("ifname", "")

    @property
    def media(self) -> str:
        """Media port is connected to."""
        return self.raw.get("media", "")

    @property
    def name(self) -> str:
        """Port name."""
        return self.raw["name"]

    @property
    def port_idx(self) -> Optional[int]:
        """Port index."""
        return self.raw.get("port_idx")

    @property
    def poe_class(self) -> str:
        """Port POE class."""
        return self.raw.get("poe_class", "")

    @property
    def poe_enable(self) -> Optional[bool]:
        """Is POE supported/requested by client."""
        return self.raw.get("poe_enable")

    @property
    def poe_mode(self) -> str:
        """Is POE auto, pasv24, passthrough, off or None."""
        return self.raw.get("poe_mode", "")

    @property
    def poe_power(self) -> str:
        """POE power usage."""
        return self.raw.get("poe_power", "")

    @property
    def poe_voltage(self) -> str:
        """POE voltage usage."""
        return self.raw.get("poe_voltage", "")

    @property
    def portconf_id(self) -> str:
        """Port configuration ID."""
        return self.raw.get("portconf_id", "")

    @property
    def port_poe(self) -> bool:
        """Is POE used."""
        return self.raw.get("port_poe") is True

    @property
    def up(self) -> str:
        """Is port up."""
        return self.raw.get("up", "")

    def __repr__(self) -> str:
        """Return the representation."""
        return f"<{self.name}: Poe {self.poe_enable}>"


class Ports:
    """Represents ports on a device."""

    def __init__(self, raw: List[dict]) -> None:
        """Initialize port manager."""
        self.ports: Dict[Union[int, str], Port] = {}
        for raw_port in raw:
            port = Port(raw_port)

            if (index := port.port_idx) is not None:
                self.ports[index] = port
            elif ifname := port.ifname:
                self.ports[ifname] = port

    def update(self, raw: List[dict]) -> None:
        """Update ports."""
        for raw_port in raw:
            index = None

            if "port_idx" in raw_port:
                index = raw_port["port_idx"]

            elif "ifname" in raw_port:
                index = raw_port["ifname"]

            if index in self.ports:
                self.ports[index].raw = raw_port

    def values(self) -> ValuesView[Port]:
        """Return ports."""
        return self.ports.values()

    def __getitem__(self, obj_id: int) -> Port:
        """Get specific port based on key."""
        return self.ports[obj_id]

    def __iter__(self) -> Iterator[Union[int, str]]:
        """Iterate over ports."""
        return iter(self.ports)



"""API management class and base class for the different end points."""

import logging
from pprint import pformat
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    ItemsView,
    Iterator,
    List,
    Optional,
    Union,
    ValuesView,
)

from .events import Event as UniFiEvent

LOGGER = logging.getLogger(__name__)

SOURCE_DATA = "data"
SOURCE_EVENT = "event"


class APIItem:
    """Base class for all end points using APIItems class."""

    def __init__(
        self,
        raw: dict,
        request: Callable[..., Awaitable[List[dict]]],
    ) -> None:
        """Initialize API item."""
        self._raw = raw
        self._request = request
        self._event: Optional[UniFiEvent] = None
        self._source = SOURCE_DATA
        self._callbacks: List[Callable] = []

    @property
    def raw(self) -> dict:
        """Read only raw data."""
        return self._raw

    @property
    def event(self) -> Optional[UniFiEvent]:
        """Read only event data."""
        return self._event

    @property
    def last_updated(self) -> str:
        """Which source, data or event last called update."""
        return self._source

    def update(
        self,
        raw: Optional[dict] = None,
        event: Optional[UniFiEvent] = None,
    ) -> None:
        """Update raw data and signal new data is available."""
        if raw:
            self._raw = raw
            self._source = SOURCE_DATA

        elif event:
            self._event = event
            self._source = SOURCE_EVENT

        else:
            return None

        for signal_update in self._callbacks:
            signal_update()

    def register_callback(self, callback: Callable) -> None:
        """Register callback for signalling.

        Callback will be used by update.
        """
        self._callbacks.append(callback)

    def remove_callback(self, callback: Callable) -> None:
        """Remove registered callback."""
        if callback in self._callbacks:
            self._callbacks.remove(callback)


class APIItems:
    """Base class for a map of API Items."""

    KEY = ""

    def __init__(
        self,
        raw: list,
        request: Callable[..., Awaitable[List[dict]]],
        path: str,
        item_cls: Any,
    ) -> None:
        """Initialize API items."""
        self._request = request
        self._path = path
        self._item_cls = item_cls
        self._items: Dict[Union[int, str], Any] = {}
        self.process_raw(raw)
        LOGGER.debug(pformat(raw))

    async def update(self) -> None:
        """Refresh data."""
        raw = await self._request("get", self._path)
        self.process_raw(raw)

    def process_raw(self, raw: List[dict]) -> set:
        """Process data."""
        new_items = set()

        for raw_item in raw:
            key = raw_item[self.KEY]

            if (obj := self._items.get(key)) is not None:
                obj.update(raw=raw_item)

            else:
                self._items[key] = self._item_cls(raw_item, self._request)
                new_items.add(key)

        return new_items

    def process_event(self, events: List[UniFiEvent]) -> set:
        """Process event."""
        new_items = set()

        for event in events:

            if (obj := self._items.get(event.mac)) is not None:
                obj.update(event=event)
                new_items.add(event.mac)

        return new_items

    def remove(self, raw: list) -> set:
        """Remove list of items."""
        removed_items = set()

        for raw_item in raw:
            key = raw_item[self.KEY]

            if key in self._items:
                self._items.pop(key)
                removed_items.add(key)

        return removed_items

    def items(self) -> ItemsView[Union[int, str], Any]:
        """Return item values."""
        return self._items.items()

    def values(self) -> ValuesView[Any]:
        """Return item values."""
        return self._items.values()

    def get(
        self,
        obj_id: Union[int, str],
        default: Optional[Any] = None,
    ) -> Optional[Any]:
        """Get item value based on key, return default if no match."""
        return self._items.get(obj_id, default)

    def __contains__(self, obj_id: Union[int, str]) -> bool:
        """Validate membership of item ID."""
        return obj_id in self._items

    def __getitem__(self, obj_id: Union[int, str]) -> Any:
        """Get item value based on key."""
        return self._items[obj_id]

    def __iter__(self) -> Iterator[Union[int, str]]:
        """Allow iterate over items."""
        return iter(self._items)



"""Setup for aioUnifi"""

from setuptools import setup

setup(
    name="aiounifi",
    packages=["aiounifi"],
    version="28",
    description="An asynchronous Python library for communicating with Unifi Controller API",
    author="Robert Svensson",
    author_email="Kane610@users.noreply.github.com",
    license="MIT",
    url="https://github.com/Kane610/aiounifi",
    download_url="https://github.com/Kane610/aiounifi/archive/v28.tar.gz",
    install_requires=["aiohttp"],
    tests_require=["pytest-asyncio", "pytest-aiohttp", "pytest", "aioresponses"],
    keywords=["unifi", "homeassistant"],
    classifiers=["Natural Language :: English", "Programming Language :: Python :: 3"],
)
